# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# Evaluation of Taylor expansions and general polynomials up to 4th order in 1D, 2D and 3D.
#
# JJ 2016-11-30

# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True

from __future__ import division
from __future__ import absolute_import

# BUG in Cython 0.20.1post0: /usr/lib/python2.7/dist-packages/Cython/Includes/libc/math.pxd
# defines fma(double x, double y), but it should be fma(double x, double y, double z)
# so we import it manually.
#
#from libc.math cimport fma   # this works in newer Cythons, the bug has been fixed
cdef extern from "<math.h>" nogil:
    double fma(double x, double y, double z)

cimport wlsqm.fitter.defs as defs  # C constants

####################################################
# Polynomial evaluation
####################################################

# These are implemented as a separate layer in order to allow us to re-use the same implementation
# for interpolating both the function value and the derivatives of the model.


# Evaluate an up to 4th order Taylor series expansion in 3D space, with its origin at (xi,yi,zi).
#
# This version uses "partially baked" coefficients, already accounting for the constant factors in the Taylor series!
# (This allows expressing the derivative information at the point xi in the raw coefficient data.)
#
# The implementation uses a symmetric Horner-like form with fused multiply-adds.
# (Note however that some of the symmetry is lost due to the way the partial results are summed.)
#
# order : in, the order of the expansion (0,1,2,3 or 4)
# fi    : in, coefficient array ("order" determines the number of entries, no separate size parameter needed)
#         The ordering of the coefficients follows the numbering of the i3_* constants in wlsqm.fitter.defs.
#         Here fi[ i3_F_c ] is f at xi, fi[ i3_X ] is df/fx at xi, fi[ i3_Y ] is df/dy at xi, fi[ i3_X2 ] is d2f/dx2 at xi, ...
# xi    : in, origin of the model, x component
# yi    : in, origin of the model, y component
# zi    : in, origin of the model, z component
# x     : in, the points where to evaluate the Taylor expansion
# out   : out, array of size (x.shape[0],); the result
#
# Return value: 0 on success, anything else indicates failure
#
cdef int taylor_3D( int order, double* fi, double xi, double yi, double zi, double[::view.generic,::view.contiguous] x, double* out ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx, dy, dz
    cdef double dxdy, dydz, dxdz
    cdef double acc1, acc2  # accumulators
    cdef double resX, resY, resZ, resXY, resYZ, resXZ, resXYZ  # intermediate results

    if order == 4:
        for k in range(n):
            dx   = x[k,0] - xi
            dy   = x[k,1] - yi
            dz   = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dz2  = dz*dz
#            dx3  = dx2*dx
#            dy3  = dy2*dy
#            dz3  = dz2*dz
#            dxdy = dx*dy
#            dydz = dy*dz
#            dxdz = dx*dz
#            out[k] = fi[ i3_F_c ] + dx*fi[ i3_X_c ] + dy*fi[ i3_Y_c ] + dz*fi[ i3_Z_c ] \
#
#                   + 0.5*dx2*fi[ i3_X2_c ] + dxdy*fi[ i3_XY_c ] + 0.5*dy2*fi[ i3_Y2_c ] + dydz*fi[ i3_YZ_c ] + 0.5*dz2*fi[ i3_Z2_c ] + dxdz*fi[ i3_XZ_c ] \
#
#                   + onesixth*dx3*fi[ i3_X3_c ] + 0.5*dx2*dy*fi[ i3_X2Y_c ] + 0.5*dx*dy2*fi[ i3_XY2_c ] + onesixth*dy3*fi[ i3_Y3_c ] \
#                   + 0.5*dy2*dz*fi[ i3_Y2Z_c ] + 0.5*dy*dz2*fi[ i3_YZ2_c ] + onesixth*dz3*fi[ i3_Z3_c ] + 0.5*dx*dz2*fi[ i3_XZ2_c ] + 0.5*dx2*dz*fi[ i3_X2Z_c ] + dx*dy*dz*fi[ i3_XYZ_c ] \
#
#                   + one24th*dx2*dx2*fi[ i3_X4_c ] + onesixth*dx3*dy*fi[ i3_X3Y_c ] + 0.25*dxdy*dxdy*fi[ i3_X2Y2_c ] + onesixth*dx*dy3*fi[ i3_XY3_c ] + one24th*dy2*dy2*fi[ i3_Y4_c ] \
#                   + onesixth*dy3*dz*fi[ i3_Y3Z_c ] + 0.25*dydz*dydz*fi[ i3_Y2Z2_c ] + onesixth*dy*dz3*fi[ i3_YZ3_c ] + one24th*dz2*dz2*fi[ i3_Z4_c ] + onesixth*dx*dz3*fi[ i3_XZ3_c ] \
#                   + 0.25*dxdz*dxdz*fi[ i3_X2Z2_c ] + onesixth*dx3*dz*fi[ i3_X3Z_c ] + 0.5*dx2*dydz*fi[ i3_X2YZ_c ] + 0.5*dy2*dxdz*fi[ i3_XY2Z_c ] + 0.5*dxdy*dz2*fi[ i3_XYZ2_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i3_F_c ] + dx*( fi[ i3_X_c ] + dx*( 0.5*( fi[ i3_X2_c ] + dy*fi[ i3_X2Y_c ] + dz*fi[ i3_X2Z_c ] + dydz*fi[ i3_X2YZ_c ] )
#                                                             + dx*( onesixth*( fi[ i3_X3_c ] + dy*fi[ i3_X3Y_c ] + dz*fi[ i3_X3Z_c ] )
#                                                                    + dx*( one24th*fi[ i3_X4_c ] ) ) ) ) \
#
#                                  + dy*( fi[ i3_Y_c ] + dy*( 0.5*( fi[ i3_Y2_c ] + dx*fi[ i3_XY2_c ] + dz*fi[ i3_Y2Z_c ] + dxdz*fi[ i3_XY2Z_c ] )
#                                                             + dy*( onesixth*( fi[ i3_Y3_c ] + dx*fi[ i3_XY3_c ] + dz*fi[ i3_Y3Z_c ] )
#                                                                    + dy*( one24th*fi[ i3_Y4_c ] ) ) ) ) \
#
#                                  + dz*( fi[ i3_Z_c ] + dz*( 0.5*( fi[ i3_Z2_c ] + dx*fi[ i3_XZ2_c ] + dy*fi[ i3_YZ2_c ] + dxdy*fi[ i3_XYZ2_c ] )
#                                                             + dz*( onesixth*( fi[ i3_Z3_c ] + dx*fi[ i3_XZ3_c ] + dy*fi[ i3_YZ3_c ] )
#                                                                    + dz*( one24th*fi[ i3_Z4_c ] ) ) ) ) \
#
#                                  + dxdy*( fi[ i3_XY_c ] + dxdy*0.25*fi[ i3_X2Y2_c ] ) \
#                                  + dydz*( fi[ i3_YZ_c ] + dydz*0.25*fi[ i3_Y2Z2_c ] ) \
#                                  + dxdz*( fi[ i3_XZ_c ] + dxdz*0.25*fi[ i3_X2Z2_c ] ) \
#
#                                  + dx*dy*dz*fi[ i3_XYZ_c ]

            # symmetric Horner-like form using fused multiply-add (fma):

            # "X" terms (the "..." in  dx*...  above)
            acc1   = fma( dy, fi[ defs.i3_X3Y_c ], fi[ defs.i3_X3_c ] )
            acc1   = fma( dz, fi[ defs.i3_X3Z_c ], acc1 )
            acc1  *= onesixth
            acc1   = fma( dx, one24th*fi[ defs.i3_X4_c ], acc1 )

            acc2   = fma( dy,   fi[ defs.i3_X2Y_c  ], fi[ defs.i3_X2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_X2Z_c  ], acc2 )
            acc2   = fma( dydz, fi[ defs.i3_X2YZ_c ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dx, acc1, acc2 )

            resX   = fma( dx, acc2, fi[ defs.i3_X_c ] )

            # "Y" terms
            acc1   = fma( dx, fi[ defs.i3_XY3_c ], fi[ defs.i3_Y3_c ] )
            acc1   = fma( dz, fi[ defs.i3_Y3Z_c ], acc1 )
            acc1  *= onesixth
            acc1   = fma( dy, one24th*fi[ defs.i3_Y4_c ], acc1 )

            acc2   = fma( dx,   fi[ defs.i3_XY2_c  ], fi[ defs.i3_Y2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_Y2Z_c  ], acc2 )
            acc2   = fma( dxdz, fi[ defs.i3_XY2Z_c ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dy, acc1, acc2 )

            resY   = fma( dy, acc2, fi[ defs.i3_Y_c ] )

            # "Z" terms
            acc1   = fma( dx, fi[ defs.i3_XZ3_c ], fi[ defs.i3_Z3_c ] )
            acc1   = fma( dy, fi[ defs.i3_YZ3_c ], acc1 )
            acc1  *= onesixth
            acc1   = fma( dz, one24th*fi[ defs.i3_Z4_c ], acc1 )

            acc2   = fma( dx,   fi[ defs.i3_XZ2_c  ], fi[ defs.i3_Z2_c ] )
            acc2   = fma( dy,   fi[ defs.i3_YZ2_c  ], acc2 )
            acc2   = fma( dxdy, fi[ defs.i3_XYZ2_c ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dz, acc1, acc2 )

            resZ   = fma( dz, acc2, fi[ defs.i3_Z_c ] )

            # "XY", "YZ", "XZ" terms
            resXY  = fma( dxdy, 0.25*fi[ defs.i3_X2Y2_c ], fi[ defs.i3_XY_c ] )
            resYZ  = fma( dydz, 0.25*fi[ defs.i3_Y2Z2_c ], fi[ defs.i3_YZ_c ] )
            resXZ  = fma( dxdz, 0.25*fi[ defs.i3_X2Z2_c ], fi[ defs.i3_XZ_c ] )

            # sum the contributions (in an approximately increasing order of magnitude)
            #
            acc1   = dx*dy*dz*fi[ defs.i3_XYZ_c ]
            acc1   = fma( dxdy,     resXY,               acc1 )
            acc1   = fma( dydz,     resYZ,               acc1 )
            acc1   = fma( dxdz,     resXZ,               acc1 )
            acc1   = fma( dx,       resX,                acc1 )
            acc1   = fma( dy,       resY,                acc1 )
            acc1   = fma( dz,       resZ,                acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 3:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dz2  = dz*dz
#            dx3  = dx2*dx
#            dy3  = dy2*dy
#            dz3  = dz2*dz
#            dxdy = dx*dy
#            dydz = dy*dz
#            dxdz = dx*dz
#            out[k] = fi[ i3_F_c ] + dx*fi[ i3_X_c ] + dy*fi[ i3_Y_c ] + dz*fi[ i3_Z_c ] \
#
#                   + 0.5*dx2*fi[ i3_X2_c ] + dxdy*fi[ i3_XY_c ] + 0.5*dy2*fi[ i3_Y2_c ] + dydz*fi[ i3_YZ_c ] + 0.5*dz2*fi[ i3_Z2_c ] + dxdz*fi[ i3_XZ_c ] \
#
#                   + onesixth*dx3*fi[ i3_X3_c ] + 0.5*dx2*dy*fi[ i3_X2Y_c ] + 0.5*dx*dy2*fi[ i3_XY2_c ] + onesixth*dy3*fi[ i3_Y3_c ] \
#                   + 0.5*dy2*dz*fi[ i3_Y2Z_c ] + 0.5*dy*dz2*fi[ i3_YZ2_c ] + onesixth*dz3*fi[ i3_Z3_c ] + 0.5*dx*dz2*fi[ i3_XZ2_c ] + 0.5*dx2*dz*fi[ i3_X2Z_c ] + dx*dy*dz*fi[ i3_XYZ_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i3_F_c ] + dx*( fi[ i3_X_c ] + dx*( 0.5*( fi[ i3_X2_c ] + dy*fi[ i3_X2Y_c ] + dz*fi[ i3_X2Z_c ] )
#                                                             + dx*( onesixth*( fi[ i3_X3_c ] ) ) ) ) \
#
#                                  + dy*( fi[ i3_Y_c ] + dy*( 0.5*( fi[ i3_Y2_c ] + dx*fi[ i3_XY2_c ] + dz*fi[ i3_Y2Z_c ] )
#                                                             + dy*( onesixth*( fi[ i3_Y3_c ] ) ) ) ) \
#
#                                  + dz*( fi[ i3_Z_c ] + dz*( 0.5*( fi[ i3_Z2_c ] + dx*fi[ i3_XZ2_c ] + dy*fi[ i3_YZ2_c ] )
#                                                             + dz*( onesixth*( fi[ i3_Z3_c ] ) ) ) ) \
#
#                                  + dxdy*( fi[ i3_XY_c ] ) \
#                                  + dydz*( fi[ i3_YZ_c ] ) \
#                                  + dxdz*( fi[ i3_XZ_c ] ) \
#
#                                  + dx*dy*dz*fi[ i3_XYZ_c ]

            # "X" terms (the "..." in  dx*...  above)
            acc2   = fma( dy,   fi[ defs.i3_X2Y_c  ], fi[ defs.i3_X2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_X2Z_c  ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dx, onesixth*fi[ defs.i3_X3_c ], acc2 )

            resX   = fma( dx, acc2, fi[ defs.i3_X_c ] )

            # "Y" terms
            acc2   = fma( dx,   fi[ defs.i3_XY2_c  ], fi[ defs.i3_Y2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_Y2Z_c  ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dy, onesixth*fi[ defs.i3_Y3_c ], acc2 )

            resY   = fma( dy, acc2, fi[ defs.i3_Y_c ] )

            # "Z" terms
            acc2   = fma( dx,   fi[ defs.i3_XZ2_c  ], fi[ defs.i3_Z2_c ] )
            acc2   = fma( dy,   fi[ defs.i3_YZ2_c  ], acc2 )
            acc2  *= 0.5
            acc2   = fma( dz, onesixth*fi[ defs.i3_Z3_c ], acc2 )

            resZ   = fma( dz, acc2, fi[ defs.i3_Z_c ] )

            # sum the contributions
            #
            acc1   = dx*dy*dz*fi[ defs.i3_XYZ_c ]
            acc1   = fma( dxdy,     fi[ defs.i3_XY_c  ], acc1 )
            acc1   = fma( dydz,     fi[ defs.i3_YZ_c  ], acc1 )
            acc1   = fma( dxdz,     fi[ defs.i3_XZ_c  ], acc1 )
            acc1   = fma( dx,       resX,                acc1 )
            acc1   = fma( dy,       resY,                acc1 )
            acc1   = fma( dz,       resZ,                acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 2:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dz2  = dz*dz
#            dxdy = dx*dy
#            dydz = dy*dz
#            dxdz = dx*dz
#            out[k] = fi[ i3_F_c ] + dx*fi[ i3_X_c ] + dy*fi[ i3_Y_c ] + dz*fi[ i3_Z_c ] \
#                   + 0.5*dx2*fi[ i3_X2_c ] + dxdy*fi[ i3_XY_c ] + 0.5*dy2*fi[ i3_Y2_c ] + dydz*fi[ i3_YZ_c ] + 0.5*dz2*fi[ i3_Z2_c ] + dxdz*fi[ i3_XZ_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i3_F_c ] + dx*( fi[ i3_X_c ] + dx*( 0.5*( fi[ i3_X2_c ] ) ) ) \
#                                  + dy*( fi[ i3_Y_c ] + dy*( 0.5*( fi[ i3_Y2_c ] ) ) ) \
#                                  + dz*( fi[ i3_Z_c ] + dz*( 0.5*( fi[ i3_Z2_c ] ) ) ) \
#
#                                  + dxdy*( fi[ i3_XY_c ] ) \
#                                  + dydz*( fi[ i3_YZ_c ] ) \
#                                  + dxdz*( fi[ i3_XZ_c ] )

            resX   = fma( dx, 0.5*fi[ defs.i3_X2_c ], fi[ defs.i3_X_c ] )
            resY   = fma( dy, 0.5*fi[ defs.i3_Y2_c ], fi[ defs.i3_Y_c ] )
            resZ   = fma( dz, 0.5*fi[ defs.i3_Z2_c ], fi[ defs.i3_Z_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i3_XY_c ]
            acc1   = fma( dydz, fi[ defs.i3_YZ_c ], acc1 )
            acc1   = fma( dxdz, fi[ defs.i3_XZ_c ], acc1 )
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1   = fma( dz,   resZ,               acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 1:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi
#            # for documentation only:
#            out[k] = fi[ i3_F_c ] + dx*fi[ i3_X_c ] + dy*fi[ i3_Y_c ] + dz*fi[ i3_Z_c ]

            # sum the contributions
            #
            acc1   = dx*fi[ defs.i3_X_c ]
            acc1   = fma( dy, fi[ defs.i3_Y_c ], acc1 )
            acc1   = fma( dz, fi[ defs.i3_Z_c ], acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i3_F_c ]

    return 0


# Same strategy as above, but for a generic polynomial (no partially baked coefficients).
#
# Now fi[ i3_F_c ] is the constant term, fi[ i3_X ] is the coefficient of (x - xi), fi[ i3_Y ] is the coefficient of (y - yi), fi[ i3_X2 ] is the coefficient of (x - xi)**2, ...
#
cdef int general_3D( int order, double* fi, double xi, double yi, double zi, double[::view.generic,::view.contiguous] x, double* out ) nogil:
    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx, dy, dz
    cdef double dxdy, dydz, dxdz
    cdef double acc1, acc2  # accumulators
    cdef double resX, resY, resZ, resXY, resYZ, resXZ, resXYZ  # intermediate results

    if order == 4:
        for k in range(n):
            dx   = x[k,0] - xi
            dy   = x[k,1] - yi
            dz   = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz

            # symmetric Horner-like form using fused multiply-add (fma):

            # "X" terms
            acc1   = fma( dy,   fi[ defs.i3_X3Y_c  ], fi[ defs.i3_X3_c ] )
            acc1   = fma( dz,   fi[ defs.i3_X3Z_c  ], acc1 )
            acc1   = fma( dx,   fi[ defs.i3_X4_c   ], acc1 )

            acc2   = fma( dy,   fi[ defs.i3_X2Y_c  ], fi[ defs.i3_X2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_X2Z_c  ], acc2 )
            acc2   = fma( dydz, fi[ defs.i3_X2YZ_c ], acc2 )
            acc2   = fma( dx,   acc1,                 acc2 )

            resX   = fma( dx, acc2, fi[ defs.i3_X_c ] )

            # "Y" terms
            acc1   = fma( dx,   fi[ defs.i3_XY3_c  ], fi[ defs.i3_Y3_c ] )
            acc1   = fma( dz,   fi[ defs.i3_Y3Z_c  ], acc1 )
            acc1   = fma( dy,   fi[ defs.i3_Y4_c   ], acc1 )

            acc2   = fma( dx,   fi[ defs.i3_XY2_c  ], fi[ defs.i3_Y2_c ] )
            acc2   = fma( dz,   fi[ defs.i3_Y2Z_c  ], acc2 )
            acc2   = fma( dxdz, fi[ defs.i3_XY2Z_c ], acc2 )
            acc2   = fma( dy,   acc1,                 acc2 )

            resY   = fma( dy, acc2, fi[ defs.i3_Y_c ] )

            # "Z" terms
            acc1   = fma( dx,   fi[ defs.i3_XZ3_c  ], fi[ defs.i3_Z3_c ] )
            acc1   = fma( dy,   fi[ defs.i3_YZ3_c  ], acc1 )
            acc1   = fma( dz,   fi[ defs.i3_Z4_c   ], acc1 )

            acc2   = fma( dx,   fi[ defs.i3_XZ2_c  ], fi[ defs.i3_Z2_c ] )
            acc2   = fma( dy,   fi[ defs.i3_YZ2_c  ], acc2 )
            acc2   = fma( dxdy, fi[ defs.i3_XYZ2_c ], acc2 )
            acc2   = fma( dz,   acc1,                 acc2 )

            resZ   = fma( dz, acc2, fi[ defs.i3_Z_c ] )

            # "XY", "YZ", "XZ" terms
            resXY  = fma( dxdy, fi[ defs.i3_X2Y2_c ], fi[ defs.i3_XY_c ] )
            resYZ  = fma( dydz, fi[ defs.i3_Y2Z2_c ], fi[ defs.i3_YZ_c ] )
            resXZ  = fma( dxdz, fi[ defs.i3_X2Z2_c ], fi[ defs.i3_XZ_c ] )

            # sum the contributions (in an approximately increasing order of magnitude)
            #
            acc1   = dx*dy*dz*fi[ defs.i3_XYZ_c ]
            acc1   = fma( dxdy,     resXY,               acc1 )
            acc1   = fma( dydz,     resYZ,               acc1 )
            acc1   = fma( dxdz,     resXZ,               acc1 )
            acc1   = fma( dx,       resX,                acc1 )
            acc1   = fma( dy,       resY,                acc1 )
            acc1   = fma( dz,       resZ,                acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 3:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz

            # "X" terms
            acc2   = fma( dy, fi[ defs.i3_X2Y_c ], fi[ defs.i3_X2_c ] )
            acc2   = fma( dz, fi[ defs.i3_X2Z_c ], acc2 )
            acc2   = fma( dx, fi[ defs.i3_X3_c  ], acc2 )

            resX   = fma( dx, acc2, fi[ defs.i3_X_c ] )

            # "Y" terms
            acc2   = fma( dx, fi[ defs.i3_XY2_c ], fi[ defs.i3_Y2_c ] )
            acc2   = fma( dz, fi[ defs.i3_Y2Z_c ], acc2 )
            acc2   = fma( dy, fi[ defs.i3_Y3_c  ], acc2 )

            resY   = fma( dy, acc2, fi[ defs.i3_Y_c ] )

            # "Z" terms
            acc2   = fma( dx, fi[ defs.i3_XZ2_c ], fi[ defs.i3_Z2_c ] )
            acc2   = fma( dy, fi[ defs.i3_YZ2_c ], acc2 )
            acc2   = fma( dz, fi[ defs.i3_Z3_c  ], acc2 )

            resZ   = fma( dz, acc2, fi[ defs.i3_Z_c ] )

            # sum the contributions
            #
            acc1   = dx*dy*dz*fi[ defs.i3_XYZ_c ]
            acc1   = fma( dxdy,     fi[ defs.i3_XY_c  ], acc1 )
            acc1   = fma( dydz,     fi[ defs.i3_YZ_c  ], acc1 )
            acc1   = fma( dxdz,     fi[ defs.i3_XZ_c  ], acc1 )
            acc1   = fma( dx,       resX,                acc1 )
            acc1   = fma( dy,       resY,                acc1 )
            acc1   = fma( dz,       resZ,                acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 2:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi
            dxdy = dx*dy
            dydz = dy*dz
            dxdz = dx*dz

            resX   = fma( dx, fi[ defs.i3_X2_c ], fi[ defs.i3_X_c ] )
            resY   = fma( dy, fi[ defs.i3_Y2_c ], fi[ defs.i3_Y_c ] )
            resZ   = fma( dz, fi[ defs.i3_Z2_c ], fi[ defs.i3_Z_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i3_XY_c ]
            acc1   = fma( dydz, fi[ defs.i3_YZ_c ], acc1 )
            acc1   = fma( dxdz, fi[ defs.i3_XZ_c ], acc1 )
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1   = fma( dz,   resZ,               acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    elif order == 1:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dz = x[k,2] - zi

            # sum the contributions
            #
            acc1   = dx*fi[ defs.i3_X_c ]
            acc1   = fma( dy, fi[ defs.i3_Y_c ], acc1 )
            acc1   = fma( dz, fi[ defs.i3_Z_c ], acc1 )
            acc1  += fi[ defs.i3_F_c ]

            out[k] = acc1

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i3_F_c ]

    return 0


# Evaluate an up to 4th order Taylor series expansion in the plane, with its origin at (xi,yi).
#
# This version uses "partially baked" coefficients, already accounting for the constant factors in the Taylor series!
# (This allows expressing the derivative information at the point xi in the raw coefficient data.)
#
# The implementation uses a symmetric Horner-like form with fused multiply-adds.
# (Note however that some of the symmetry is lost due to the way the partial results are summed.)
#
# order : in, the order of the expansion (0,1,2,3 or 4)
# fi    : in, coefficient array ("order" determines the number of entries, no separate size parameter needed)
#         The ordering of the coefficients follows the numbering of the i2_* constants in wlsqm.fitter.defs.
#         Here fi[ i2_F_c ] is f at xi, fi[ i2_X ] is df/fx at xi, fi[ i2_Y ] is df/dy at xi, fi[ i2_X2 ] is d2f/dx2 at xi, ...
# xi    : in, origin of the model, x component
# yi    : in, origin of the model, y component
# x     : in, the points where to evaluate the Taylor expansion
# out   : out, array of size (x.shape[0],); the result
#
# Return value: 0 on success, anything else indicates failure
#
cdef int taylor_2D( int order, double* fi, double xi, double yi, double[::view.generic,::view.contiguous] x, double* out ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx, dy
    cdef double dxdy
    cdef double acc1, acc2  # accumulators
    cdef double resX, resY, resXY  # intermediate results

    if order == 4:
        for k in range(n):
            dx   = x[k,0] - xi
            dy   = x[k,1] - yi
            dxdy = dx*dy
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dx3  = dx2*dx
#            dy3  = dy2*dy
#            dxdy = dx*dy
#            out[k] = fi[ i2_F_c ] + dx*fi[ i2_X_c ] + dy*fi[ i2_Y_c ] \
#
#                   + 0.5*dx2*fi[ i2_X2_c ] + dxdy*fi[ i2_XY_c ] + 0.5*dy2*fi[ i2_Y2_c ] \
#
#                   + onesixth*dx3*fi[ i2_X3_c ] + 0.5*dx2*dy*fi[ i2_X2Y_c ] + 0.5*dx*dy2*fi[ i2_XY2_c ] + onesixth*dy3*fi[ i2_Y3_c ] \
#
#                   + one24th*dx2*dx2*fi[ i2_X4_c ] + onesixth*dx3*dy*fi[ i2_X3Y_c ] + 0.25*dxdy*dxdy*fi[ i2_X2Y2_c ] + onesixth*dx*dy3*fi[ i2_XY3_c ] + one24th*dy2*dy2*fi[ i2_Y4_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i2_F_c ] + dx*( fi[ i2_X_c ] + dx*( 0.5*( fi[ i2_X2_c ] + dy*fi[ i2_X2Y_c ] )
#                                                             + dx*( onesixth*( fi[ i2_X3_c ] + dy*fi[ i2_X3Y_c ] )
#                                                                    + dx*( one24th*fi[ i2_X4_c ] ) ) ) ) \
#
#                                  + dy*( fi[ i2_Y_c ] + dy*( 0.5*( fi[ i2_Y2_c ] + dx*fi[ i2_XY2_c ] )
#                                                             + dy*( onesixth*( fi[ i2_Y3_c ] + dx*fi[ i2_XY3_c ] )
#                                                                    + dy*( one24th*fi[ i2_Y4_c ] ) ) ) ) \
#
#                                  + dxdy*( fi[ i2_XY_c ] + dxdy*0.25*fi[ i2_X2Y2_c ] )

            # symmetric Horner-like form using fused multiply-add (fma):

            # "X" terms (the "..." in  dx*...  above)
            acc1   = fma( dy, fi[ defs.i2_X3Y_c ], fi[ defs.i2_X3_c ] )
            acc1  *= onesixth
            acc1   = fma( dx, one24th*fi[ defs.i2_X4_c ], acc1 )

            acc2   = fma( dy,   fi[ defs.i2_X2Y_c  ], fi[ defs.i2_X2_c ] )
            acc2  *= 0.5
            acc2   = fma( dx, acc1, acc2 )

            resX   = fma( dx, acc2, fi[ defs.i2_X_c ] )

            # "Y" terms
            acc1   = fma( dx, fi[ defs.i2_XY3_c ], fi[ defs.i2_Y3_c ] )
            acc1  *= onesixth
            acc1   = fma( dy, one24th*fi[ defs.i2_Y4_c ], acc1 )

            acc2   = fma( dx,   fi[ defs.i2_XY2_c  ], fi[ defs.i2_Y2_c ] )
            acc2  *= 0.5
            acc2   = fma( dy, acc1, acc2 )

            resY   = fma( dy, acc2, fi[ defs.i2_Y_c ] )

            # "XY" terms
            resXY  = fma( dxdy, 0.25*fi[ defs.i2_X2Y2_c ], fi[ defs.i2_XY_c ] )

            # sum the contributions (in an approximately increasing order of magnitude)
            #
            acc1   = dxdy*resXY
            acc1   = fma( dx,   resX,  acc1 )
            acc1   = fma( dy,   resY,  acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 3:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dxdy = dx*dy
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dx3  = dx2*dx
#            dy3  = dy2*dy
#            dxdy = dx*dy
#            out[k] = fi[ i2_F_c ] + dx*fi[ i2_X_c ] + dy*fi[ i2_Y_c ] \
#
#                   + 0.5*dx2*fi[ i2_X2_c ] + dxdy*fi[ i2_XY_c ] + 0.5*dy2*fi[ i2_Y2_c ] \
#
#                   + onesixth*dx3*fi[ i2_X3_c ] + 0.5*dx2*dy*fi[ i2_X2Y_c ] + 0.5*dx*dy2*fi[ i2_XY2_c ] + onesixth*dy3*fi[ i2_Y3_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i2_F_c ] + dx*( fi[ i2_X_c ] + dx*( 0.5*( fi[ i2_X2_c ] + dy*fi[ i2_X2Y_c ] )
#                                                             + dx*( onesixth*( fi[ i2_X3_c ] ) ) ) ) \
#
#                                  + dy*( fi[ i2_Y_c ] + dy*( 0.5*( fi[ i2_Y2_c ] + dx*fi[ i2_XY2_c ] )
#                                                             + dy*( onesixth*( fi[ i2_Y3_c ] ) ) ) ) \
#
#                                  + dxdy*( fi[ i2_XY_c ] )

            # "X" terms (the "..." in  dx*...  above)
            acc2   = fma( dy,   fi[ defs.i2_X2Y_c  ], fi[ defs.i2_X2_c ] )
            acc2  *= 0.5
            acc2   = fma( dx, onesixth*fi[ defs.i2_X3_c ], acc2 )

            resX   = fma( dx, acc2, fi[ defs.i2_X_c ] )

            # "Y" terms
            acc2   = fma( dx,   fi[ defs.i2_XY2_c  ], fi[ defs.i2_Y2_c ] )
            acc2  *= 0.5
            acc2   = fma( dy, onesixth*fi[ defs.i2_Y3_c ], acc2 )

            resY   = fma( dy, acc2, fi[ defs.i2_Y_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i2_XY_c ]
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 2:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dxdy = dx*dy
#            # for documentation only:
#
#            # naive form:
#            dx2  = dx*dx
#            dy2  = dy*dy
#            dxdy = dx*dy
#            out[k] = fi[ i2_F_c ] + dx*fi[ i2_X_c ] + dy*fi[ i2_Y_c ] \
#                   + 0.5*dx2*fi[ i2_X2_c ] + dxdy*fi[ i2_XY_c ] + 0.5*dy2*fi[ i2_Y2_c ]
#
#            # symmetric Horner-like form:
#            out[k] = fi[ i2_F_c ] + dx*( fi[ i2_X_c ] + dx*( 0.5*( fi[ i2_X2_c ] ) ) ) \
#                                  + dy*( fi[ i2_Y_c ] + dy*( 0.5*( fi[ i2_Y2_c ] ) ) ) \
#
#                                  + dxdy*( fi[ i2_XY_c ] )

            resX   = fma( dx, 0.5*fi[ defs.i2_X2_c ], fi[ defs.i2_X_c ] )
            resY   = fma( dy, 0.5*fi[ defs.i2_Y2_c ], fi[ defs.i2_Y_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i2_XY_c ]
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 1:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
#            # for documentation only:
#            out[k] = fi[ i2_F_c ] + dx*fi[ i2_X_c ] + dy*fi[ i2_Y_c ]

            # sum the contributions
            #
            acc1   = dx*fi[ defs.i2_X_c ]
            acc1   = fma( dy, fi[ defs.i2_Y_c ], acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i2_F_c ]

    return 0


# Same strategy as above, but for a generic polynomial (no partially baked coefficients).
#
# Now fi[ i2_F_c ] is the constant term, fi[ i2_X ] is the coefficient of (x - xi), fi[ i2_Y ] is the coefficient of (y - yi), fi[ i2_X2 ] is the coefficient of (x - xi)**2, ...
#
cdef int general_2D( int order, double* fi, double xi, double yi, double[::view.generic,::view.contiguous] x, double* out ) nogil:
    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx, dy
    cdef double dxdy
    cdef double acc1, acc2  # accumulators
    cdef double resX, resY, resXY  # intermediate results

    if order == 4:
        for k in range(n):
            dx   = x[k,0] - xi
            dy   = x[k,1] - yi
            dxdy = dx*dy

            # symmetric Horner-like form using fused multiply-add (fma):

            # "X" terms
            acc1   = fma( dy, fi[ defs.i2_X3Y_c ], fi[ defs.i2_X3_c ] )
            acc1   = fma( dx, fi[ defs.i2_X4_c  ], acc1 )

            acc2   = fma( dy, fi[ defs.i2_X2Y_c ], fi[ defs.i2_X2_c ] )
            acc2   = fma( dx, acc1, acc2 )

            resX   = fma( dx, acc2, fi[ defs.i2_X_c ] )

            # "Y" terms
            acc1   = fma( dx, fi[ defs.i2_XY3_c ], fi[ defs.i2_Y3_c ] )
            acc1   = fma( dy, fi[ defs.i2_Y4_c  ], acc1 )

            acc2   = fma( dx, fi[ defs.i2_XY2_c ], fi[ defs.i2_Y2_c ] )
            acc2   = fma( dy, acc1, acc2 )

            resY   = fma( dy, acc2, fi[ defs.i2_Y_c ] )

            # "XY" terms
            resXY  = fma( dxdy, fi[ defs.i2_X2Y2_c ], fi[ defs.i2_XY_c ] )

            # sum the contributions
            #
            acc1   = dxdy*resXY
            acc1   = fma( dx,   resX,  acc1 )
            acc1   = fma( dy,   resY,  acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 3:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dxdy = dx*dy

            # "X" terms
            acc2   = fma( dy, fi[ defs.i2_X2Y_c ], fi[ defs.i2_X2_c ] )
            acc2   = fma( dx, fi[ defs.i2_X3_c  ], acc2 )

            resX   = fma( dx, acc2, fi[ defs.i2_X_c ] )

            # "Y" terms
            acc2   = fma( dx, fi[ defs.i2_XY2_c ], fi[ defs.i2_Y2_c ] )
            acc2   = fma( dy, fi[ defs.i2_Y3_c  ], acc2 )

            resY   = fma( dy, acc2, fi[ defs.i2_Y_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i2_XY_c ]
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 2:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
            dxdy = dx*dy

            resX   = fma( dx, fi[ defs.i2_X2_c ], fi[ defs.i2_X_c ] )
            resY   = fma( dy, fi[ defs.i2_Y2_c ], fi[ defs.i2_Y_c ] )

            # sum the contributions
            #
            acc1   = dxdy*fi[ defs.i2_XY_c ]
            acc1   = fma( dx,   resX,               acc1 )
            acc1   = fma( dy,   resY,               acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    elif order == 1:
        for k in range(n):
            dx = x[k,0] - xi
            dy = x[k,1] - yi
#            # for documentation only:
#            out[k] = fi[ i2_F_c ] + dx*fi[ i2_X_c ] + dy*fi[ i2_Y_c ]

            # sum the contributions
            #
            acc1   = dx*fi[ defs.i2_X_c ]
            acc1   = fma( dy, fi[ defs.i2_Y_c ], acc1 )
            acc1  += fi[ defs.i2_F_c ]

            out[k] = acc1

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i2_F_c ]

    return 0


# Evaluate an up to 4th order Taylor series expansion on the real axis, with its origin at (xi).
#
# The implementation uses Horner's form with fused multiply-adds.
#
# order : in, the order of the expansion (0,1,2,3 or 4)
# fi    : in, coefficient array ("order" determines the number of entries, no separate size parameter needed)
#         The ordering of the coefficients follows the numbering of the i1_* constants in wlsqm.fitter.defs.
#         Here fi[ i1_F_c ] is f at xi, fi[ i1_X ] is df/dx at xi, fi[ i1_X2 ] is d2f/dx2 at xi, ...
# xi    : in, origin of the model
# x     : in, the points where to evaluate the Taylor expansion
# out   : out, array of size (x.shape[0],); the result
#
# Return value: 0 on success, anything else indicates failure
#
cdef int taylor_1D( int order, double* fi, double xi, double[::view.generic] x, double* out ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx
    cdef double acc

    if order == 4:
        for k in range(n):
            dx = x[k] - xi
#            # for documentation only:
#
#            # naive form:
#            dx2 = dx*dx
#            out[k] = fi[ i1_F_c ] + dx*fi[ i1_X_c ] + 0.5*dx2*fi[ i1_X2_c ] + onesixth*dx2*dx*fi[ i1_X3_c ] + one24th*dx2*dx2*fi[ i1_X4_c ]
#
#            # Horner form:
#            out[k] = fi[ i1_F_c ] + dx*( fi[ i1_X_c ] + dx*( 0.5*fi[ i1_X2_c ] + dx*( onesixth*fi[ i1_X3_c ] + dx*( one24th*fi[ i1_X4_c ] ) ) ) )

            # Horner form using fused multiply-add:
            acc    = fma( dx, one24th*fi[ defs.i1_X4_c ], onesixth*fi[ defs.i1_X3_c ] )
            acc    = fma( dx, acc, 0.5*fi[ defs.i1_X2_c ] )
            acc    = fma( dx, acc, fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 3:
        for k in range(n):
            dx = x[k] - xi
#            # for documentation only:
#
#            # naive form:
#            dx2 = dx*dx
#            out[k] = fi[ i1_F_c ] + dx*fi[ i1_X_c ] + 0.5*dx2*fi[ i1_X2_c ] + onesixth*dx2*dx*fi[ i1_X3_c ]
#
#            # Horner form:
#            out[k] = fi[ i1_F_c ] + dx*( fi[ i1_X_c ] + dx*( 0.5*fi[ i1_X2_c ] + dx*( onesixth*fi[ i1_X3_c ] ) ) )

            # Horner form using fused multiply-add:
            acc    = fma( dx, onesixth*fi[ defs.i1_X3_c ], 0.5*fi[ defs.i1_X2_c ] )
            acc    = fma( dx, acc, fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 2:
        for k in range(n):
            dx  = x[k] - xi
#            # for documentation only:
#
#            # naive form:
#            out[k] = fi[ i1_F_c ] + dx*fi[ i1_X_c ] + 0.5*dx*dx*fi[ i1_X2_c ]
#
#            # Horner form:
#            out[k] = fi[ i1_F_c ] + dx*( fi[ i1_X_c ] + dx*( 0.5*fi[ i1_X2_c ] ) )

            # Horner form using fused multiply-add:
            acc    = fma( dx, 0.5*fi[ defs.i1_X2_c ], fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 1:
        for k in range(n):
            dx  = x[k] - xi
#            out[k] = fi[ i1_F_c ] + dx*fi[ i1_X_c ]

            # Horner form using fused multiply-add:
            out[k] = fma( dx, fi[ defs.i1_X_c ], fi[ defs.i1_F_c ] )  # fused multiply-add,  op1*op2 + op3

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i1_F_c ]

    return 0


# Same strategy as above, but for a generic polynomial (no partially baked coefficients).
#
# Now fi[ i1_F_c ] is the constant term, fi[ i1_X ] is the coefficient of (x - xi), fi[ i2_X2 ] is the coefficient of (x - xi)**2, ...
#
cdef int general_1D( int order, double* fi, double xi, double[::view.generic] x, double* out ) nogil:
    if order not in [0,1,2,3,4]:
        return -1
#        with gil:
#            raise ValueError( "order must be 1, 2, 3 or 4; got %d" % order )

    cdef int n = x.shape[0]
    cdef int k
    cdef double dx
    cdef double acc  # accumulator

    if order == 4:
        for k in range(n):
            dx = x[k] - xi

            acc    = fma( dx, fi[ defs.i1_X4_c ], fi[ defs.i1_X3_c ] )
            acc    = fma( dx, acc, fi[ defs.i1_X2_c ] )
            acc    = fma( dx, acc, fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 3:
        for k in range(n):
            dx = x[k] - xi

            acc    = fma( dx, fi[ defs.i1_X3_c ], fi[ defs.i1_X2_c ] )
            acc    = fma( dx, acc, fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 2:
        for k in range(n):
            dx = x[k] - xi

            acc    = fma( dx, fi[ defs.i1_X2_c ], fi[ defs.i1_X_c ] )
            out[k] = fma( dx, acc, fi[ defs.i1_F_c ] )

    elif order == 1:
        for k in range(n):
            dx = x[k] - xi

            out[k] = fma( dx, fi[ defs.i1_X_c ], fi[ defs.i1_F_c ] )  # fused multiply-add,  op1*op2 + op3

    else: # order == 0:
        for k in range(n):
            out[k] = fi[ defs.i1_F_c ]

    return 0




# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# Cython declarations for the main module. See the .pyx source for wlsqm.fitter.simple for documentation.
#
# JJ 2016-11-07

# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True

# This module contains "driver" routines in the LAPACK sense.
# The low-level C routines are contained in wlsqm.fitter.impl.

from cython cimport view  # for usage, see http://cython.readthedocs.io/en/latest/src/userguide/memoryviews.html#specifying-more-general-memory-layouts

####################################################
# Single case (one neighborhood), single-threaded
####################################################

cdef int generic_fit_basic( int dimension, double[::view.generic,::view.contiguous] xkManyD, double[::view.generic] xk1D, double[::view.generic] fk, double[::1] xiManyD, double xi1D, double[::1] fi,
                                           double[::view.generic,::view.contiguous] sens, int do_sens, int order, long long knowns, int weighting_method, int debug ) nogil except -1

cdef int generic_fit_iterative( int dimension, double[::view.generic,::view.contiguous] xkManyD, double[::view.generic] xk1D, double[::view.generic] fk, double[::1] xiManyD, double xi1D, double[::1] fi,
                                               double[::view.generic,::view.contiguous] sens, int do_sens, int order, long long knowns, int weighting_method, int max_iter, int debug ) nogil except -1

####################################################
# Many cases, single-threaded
####################################################

# single-threaded
cdef int generic_fit_basic_many( int dimension, double[::view.generic,::view.generic,::view.contiguous] xkManyD, double[::view.generic,::view.generic] xk1D,
                                                double[::view.generic,::view.generic] fk, int[::view.generic] nk,
                                                double[::view.generic,::view.contiguous] xiManyD, double[::view.generic] xi1D, double[::view.generic,::view.contiguous] fi,
                                                double[::view.generic,::view.generic,::view.contiguous] sens, int do_sens,
                                                int[::view.generic] order, long long[::view.generic] knowns, int[::view.generic] weighting_method, int debug ) nogil except -1

cdef int generic_fit_iterative_many( int dimension, double[::view.generic,::view.generic,::view.contiguous] xkManyD, double[::view.generic,::view.generic] xk1D,
                                                    double[::view.generic,::view.generic] fk, int[::view.generic] nk,
                                                    double[::view.generic,::view.contiguous] xiManyD, double[::view.generic] xi1D, double[::view.generic,::view.contiguous] fi,
                                                    double[::view.generic,::view.generic,::view.contiguous] sens, int do_sens,
                                                    int[::view.generic] order, long long[::view.generic] knowns, int[::view.generic] weighting_method, int max_iter, int debug ) nogil except -1

####################################################
# Many cases, multithreaded
####################################################

cdef int generic_fit_basic_many_parallel( int dimension, double[::view.generic,::view.generic,::view.contiguous] xkManyD, double[::view.generic,::view.generic] xk1D,
                                                         double[::view.generic,::view.generic] fk, int[::view.generic] nk,
                                                         double[::view.generic,::view.contiguous] xiManyD, double[::view.generic] xi1D, double[::view.generic,::view.contiguous] fi,
                                                         double[::view.generic,::view.generic,::view.contiguous] sens, int do_sens,
                                                         int[::view.generic] order, long long[::view.generic] knowns, int[::view.generic] weighting_method, int ntasks, int debug ) nogil except -1

cdef int generic_fit_iterative_many_parallel( int dimension, double[::view.generic,::view.generic,::view.contiguous] xkManyD, double[::view.generic,::view.generic] xk1D,
                                                             double[::view.generic,::view.generic] fk, int[::view.generic] nk,
                                                             double[::view.generic,::view.contiguous] xiManyD, double[::view.generic] xi1D, double[::view.generic,::view.contiguous] fi,
                                                             double[::view.generic,::view.generic,::view.contiguous] sens, int do_sens,
                                                             int[::view.generic] order, long long[::view.generic] knowns, int[::view.generic] weighting_method, int max_iter, int ntasks, int debug ) nogil except -1




# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# C-level definitions for Cython.
#
# This file contains only the declarations; the actual values are set in the .pyx source for wlsqm.fitter.defs.
#
# The suffix of _c means "visible at the C level in Cython"; it is used to distinguish
# the typed C constants from the Python-accessible objects also defined in the .pyx source.
#
# JJ 2016-11-30

# Algorithms for the solve step (expert mode).
#
cdef int ALGO_BASIC_c      # fit just once
cdef int ALGO_ITERATIVE_c  # fit with iterative refinement to mitigate roundoff

# Weighting methods.
#
cdef int WEIGHT_UNIFORM_c
cdef int WEIGHT_CENTER_c

# DOF index in the array f.
#
# These are ordered in increasing order of number of differentiations, so that if only first derivatives
# are required, the DOF array can be simply truncated after the first derivatives.
#
# To avoid gaps in the numbering, this requires separate DOF orderings for the 1D, 2D and 3D cases.
#
# (The other logical possibility would be function value first, then x-related, then y-related, then mixed,
#  but then the case of "first derivatives only" requires changes to the ordering to avoid gaps.
#  Specifying different orderings for different numbers of space dimensions is conceptually cleaner
#  of the two possibilities.)

# 1D case
#
cdef int i1_F_c
cdef int i1_X_c
cdef int i1_X2_c
cdef int i1_X3_c
cdef int i1_X4_c

cdef int i1_0th_end_c  # one-past end of zeroth-order case
cdef int i1_1st_end_c  # one-past end of first-order case
cdef int i1_2nd_end_c  # one-past-end of second-order case
cdef int i1_3rd_end_c  # one-past-end of third-order case
cdef int i1_4th_end_c  # one-past-end of fourth-order case

cdef int SIZE1_c  # maximum possible number of DOFs, 1D case

# 2D case
#
cdef int i2_F_c

cdef int i2_X_c
cdef int i2_Y_c

cdef int i2_X2_c
cdef int i2_XY_c
cdef int i2_Y2_c

cdef int i2_X3_c
cdef int i2_X2Y_c
cdef int i2_XY2_c
cdef int i2_Y3_c

cdef int i2_X4_c
cdef int i2_X3Y_c
cdef int i2_X2Y2_c
cdef int i2_XY3_c
cdef int i2_Y4_c

cdef int i2_0th_end_c  # one-past end of zeroth-order case
cdef int i2_1st_end_c  # one-past end of first-order case
cdef int i2_2nd_end_c  # one-past-end of second-order case
cdef int i2_3rd_end_c  # one-past-end of third-order case
cdef int i2_4th_end_c  # one-past-end of fourth-order case

cdef int SIZE2_c  # maximum possible number of DOFs, 2D case

# 3D case
#
cdef int i3_F_c

cdef int i3_X_c
cdef int i3_Y_c
cdef int i3_Z_c

cdef int i3_X2_c
cdef int i3_XY_c
cdef int i3_Y2_c
cdef int i3_YZ_c
cdef int i3_Z2_c
cdef int i3_XZ_c

cdef int i3_X3_c
cdef int i3_X2Y_c
cdef int i3_XY2_c
cdef int i3_Y3_c
cdef int i3_Y2Z_c
cdef int i3_YZ2_c
cdef int i3_Z3_c
cdef int i3_XZ2_c
cdef int i3_X2Z_c
cdef int i3_XYZ_c

cdef int i3_X4_c
cdef int i3_X3Y_c
cdef int i3_X2Y2_c
cdef int i3_XY3_c
cdef int i3_Y4_c
cdef int i3_Y3Z_c
cdef int i3_Y2Z2_c
cdef int i3_YZ3_c
cdef int i3_Z4_c
cdef int i3_XZ3_c
cdef int i3_X2Z2_c
cdef int i3_X3Z_c
cdef int i3_X2YZ_c
cdef int i3_XY2Z_c
cdef int i3_XYZ2_c

cdef int i3_0th_end_c  # one-past end of zeroth-order case
cdef int i3_1st_end_c  # one-past end of first-order case
cdef int i3_2nd_end_c  # one-past-end of second-order case
cdef int i3_3rd_end_c  # one-past-end of third-order case
cdef int i3_4th_end_c  # one-past-end of fourth-order case

cdef int SIZE3_c  # maximum possible number of DOFs, 3D case


# bitmask constants for knowns.
#
# Knowns are eliminated algebraically from the equation system; if any knowns are specified,
# the system to be solved (for a point x_i) will be smaller than the full 6x6.
#
# The sensible default thing to do is to consider the function value F known, with all the
# derivatives unknown.
#
# Note that here "known" means "known at point x_i" (the point at which we wish to compute the derivatives).
#
# Function values (F) are always assumed known at all *neighbor* points x_k, since they are used
# for determining the local least-squares quadratic polynomial fit to the data. This fit is then used
# as local a surrogate model for the unknown function f; in WLSQM, the derivatives are actually computed
# from the surrogate.
#
# The option to have the function value (F) as an unknown is useful with Neumann BCs, if the neighborhoods
# of the Neumann boundary points are chosen so that each Neumann boundary point only uses neighbors from
# the interior of the domain. This gives the possibility to leave F free at all Neumann boundary points,
# while prescribing only a derivative.
#
# (In practice, at slanted (i.e. not coordinate axis aligned) boundaries, local (tangent, normal)
#  coordinates must be used; i.e., the coordinate system in which the derivatives are to be computed
#  must be rotated to match the orientation of the boundary. This makes Y the normal derivative,
#  which can then be prescribed using this mechanism, while leaving the function value F free.)

# 1D case
#
cdef long long b1_F_c
cdef long long b1_X_c
cdef long long b1_X2_c
cdef long long b1_X3_c
cdef long long b1_X4_c

# 2D case
#
cdef long long b2_F_c

cdef long long b2_X_c
cdef long long b2_Y_c

cdef long long b2_X2_c
cdef long long b2_XY_c
cdef long long b2_Y2_c

cdef long long b2_X3_c
cdef long long b2_X2Y_c
cdef long long b2_XY2_c
cdef long long b2_Y3_c

cdef long long b2_X4_c
cdef long long b2_X3Y_c
cdef long long b2_X2Y2_c
cdef long long b2_XY3_c
cdef long long b2_Y4_c

# 3D case
#
cdef long long b3_F_c

cdef long long b3_X_c
cdef long long b3_Y_c
cdef long long b3_Z_c

cdef long long b3_X2_c
cdef long long b3_XY_c
cdef long long b3_Y2_c
cdef long long b3_YZ_c
cdef long long b3_Z2_c
cdef long long b3_XZ_c

cdef long long b3_X3_c
cdef long long b3_X2Y_c
cdef long long b3_XY2_c
cdef long long b3_Y3_c
cdef long long b3_Y2Z_c
cdef long long b3_YZ2_c
cdef long long b3_Z3_c
cdef long long b3_XZ2_c
cdef long long b3_X2Z_c
cdef long long b3_XYZ_c

cdef long long b3_X4_c
cdef long long b3_X3Y_c
cdef long long b3_X2Y2_c
cdef long long b3_XY3_c
cdef long long b3_Y4_c
cdef long long b3_Y3Z_c
cdef long long b3_Y2Z2_c
cdef long long b3_YZ3_c
cdef long long b3_Z4_c
cdef long long b3_XZ3_c
cdef long long b3_X2Z2_c
cdef long long b3_X3Z_c
cdef long long b3_X2YZ_c
cdef long long b3_XY2Z_c
cdef long long b3_XYZ2_c




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.

This module contains C-level and Python-level definitions of constants. The constants are made visible to Python by creating Python objects, with their values copied from the corresponding C constants.

In the source code, the suffix of _c means "visible at the C level in Cython"; it is used to distinguish the typed C constants from the corresponding Python objects.

Naming scheme:

    ALGO_*   = algorithms for the solve step (for advanced API in wlsqm.fitter.expert).
    WEIGHT_* = weighting methods for the error (data - predicted) in least-squares fitting.

    i1_* = integer, 1D case
    i2_* = integer, 2D case
    i3_* = integer, 3D case

        The i?_* constants are the human-readable names for the DOF indices in the "fi" array (see docstrings in wlsqm.fitter.simple).

    b1_* = bitmask, 1D case
    b2_* = bitmask, 2D case
    b3_* = bitmask, 3D case

    *_end = one-past-end index of this case. The ordinal (0th, 1st, 2nd, 3rd, 4th) refers to the degree of the fit.
            E.g. i2_3rd_end = one-past-end for 2D with 3rd order fit.

    SIZE1 = maximum possible number of DOFs (degrees of freedom), 1D case
    SIZE2 = maximum possible number of DOFs (degrees of freedom), 2D case
    SIZE3 = maximum possible number of DOFs (degrees of freedom), 3D case

        ("maximum possible" because if order < 4, then only lower-degree DOFs will exist.)

    F  = function value
    X  = "times x" (coefficients) or "differentiate by x" (see wlsqm.fitter.interp)
    X2 = "times x**2" or "differentiate twice by x"
    Y, Z respectively

Examples:

    i2_F   = 2D case, function value
    i2_X2Y = 2D case, coefficient of the X**2 * Y term in the polynomial; or request differentiation twice by x and once by y to compute d3f/dx2dy (see wlsqm.fitter.interp)

        IMPORTANT: the DOF values returned by the fitter are "partially baked" such that the DOF value directly corresponds to the value of the corresponding derivative.
                   This is for convenience of evaluating derivatives at the model reference point.

                   E.g. fi[:,i2_X2] is the coefficient of d2f/dx2 in a Taylor series expansion of f around the reference point xi.
                   (The ":" is here meant to refer to the reference point xi for all local models; see  wlsqm.fitter.simple.fit_2D_many()  for a description of the "fi" array.)

JJ 2016-11-30
"""

#################################################
# C definitions (Cython level)
#################################################

# Algorithms for the solve step (expert mode).
#
cdef int ALGO_BASIC_c     = 1  # just fit once
cdef int ALGO_ITERATIVE_c = 2  # fit with iterative refinement to mitigate roundoff

# Weighting methods.
#
cdef int WEIGHT_UNIFORM_c = 1
cdef int WEIGHT_CENTER_c  = 2

# DOF index in the array f.
#
# These are ordered in increasing order of number of differentiations, so that if only first derivatives
# are required, the DOF array can be simply truncated after the first derivatives.
#
# To avoid gaps in the numbering, this requires separate DOF orderings for the 1D, 2D and 3D cases.
#
# (The other logical possibility would be function value first, then x-related, then y-related, then mixed,
#  but then the case of "first derivatives only" requires changes to the ordering to avoid gaps.
#  Specifying different orderings for different numbers of space dimensions is conceptually cleaner
#  of the two possibilities.)

# 1D case
#
cdef int i1_F_c   = 0
cdef int i1_X_c   = 1
cdef int i1_X2_c  = 2
cdef int i1_X3_c  = 3
cdef int i1_X4_c  = 4

cdef int i1_0th_end_c = 1  # one-past end of zeroth-order case
cdef int i1_1st_end_c = 2  # one-past end of first-order case
cdef int i1_2nd_end_c = 3  # one-past-end of second-order case
cdef int i1_3rd_end_c = 4  # one-past-end of third-order case
cdef int i1_4th_end_c = 5  # one-past-end of fourth-order case

cdef int SIZE1_c = i1_4th_end_c  # maximum possible number of DOFs, 1D case

# 2D case
#
cdef int i2_F_c     =  0

cdef int i2_X_c     =  1
cdef int i2_Y_c     =  2

cdef int i2_X2_c    =  3
cdef int i2_XY_c    =  4
cdef int i2_Y2_c    =  5

cdef int i2_X3_c    =  6
cdef int i2_X2Y_c   =  7
cdef int i2_XY2_c   =  8
cdef int i2_Y3_c    =  9

cdef int i2_X4_c    = 10
cdef int i2_X3Y_c   = 11
cdef int i2_X2Y2_c  = 12
cdef int i2_XY3_c   = 13
cdef int i2_Y4_c    = 14

cdef int i2_0th_end_c =  1  # one-past end of zeroth-order case
cdef int i2_1st_end_c =  3  # one-past end of first-order case
cdef int i2_2nd_end_c =  6  # one-past-end of second-order case
cdef int i2_3rd_end_c = 10  # one-past-end of third-order case
cdef int i2_4th_end_c = 15  # one-past-end of fourth-order case

cdef int SIZE2_c = i2_4th_end_c  # maximum possible number of DOFs, 2D case

# 3D case
#
cdef int i3_F_c      =  0

cdef int i3_X_c      =  1
cdef int i3_Y_c      =  2
cdef int i3_Z_c      =  3

cdef int i3_X2_c     =  4
cdef int i3_XY_c     =  5
cdef int i3_Y2_c     =  6
cdef int i3_YZ_c     =  7
cdef int i3_Z2_c     =  8
cdef int i3_XZ_c     =  9

cdef int i3_X3_c     = 10
cdef int i3_X2Y_c    = 11
cdef int i3_XY2_c    = 12
cdef int i3_Y3_c     = 13
cdef int i3_Y2Z_c    = 14
cdef int i3_YZ2_c    = 15
cdef int i3_Z3_c     = 16
cdef int i3_XZ2_c    = 17
cdef int i3_X2Z_c    = 18
cdef int i3_XYZ_c    = 19

cdef int i3_X4_c     = 20
cdef int i3_X3Y_c    = 21
cdef int i3_X2Y2_c   = 22
cdef int i3_XY3_c    = 23
cdef int i3_Y4_c     = 24
cdef int i3_Y3Z_c    = 25
cdef int i3_Y2Z2_c   = 26
cdef int i3_YZ3_c    = 27
cdef int i3_Z4_c     = 28
cdef int i3_XZ3_c    = 29
cdef int i3_X2Z2_c   = 30
cdef int i3_X3Z_c    = 31
cdef int i3_X2YZ_c   = 32
cdef int i3_XY2Z_c   = 33
cdef int i3_XYZ2_c   = 34

cdef int i3_0th_end_c =  1  # one-past end of zeroth-order case
cdef int i3_1st_end_c =  4  # one-past end of first-order case
cdef int i3_2nd_end_c = 10  # one-past-end of second-order case
cdef int i3_3rd_end_c = 20  # one-past-end of third-order case
cdef int i3_4th_end_c = 35  # one-past-end of fourth-order case

cdef int SIZE3_c = i3_4th_end_c  # maximum possible number of DOFs, 3D case


# bitmask constants for knowns.
#
# Knowns are eliminated algebraically from the equation system; if any knowns are specified,
# the system to be solved (for a point x_i) will be smaller than the full system (e.g. "full" is 6x6 for 2nd order in 2D).
#
# The sensible default is to consider the function value F known, with all the derivatives unknown.
#
# Note that here "known" means "known at point xi" (the reference point of the model).
#
# Function values (F) are always assumed known at all *neighbor* points xk, since they are used
# for determining the local least-squares polynomial fit to the data. This fit is then used
# as a local surrogate model representing the unknown function f.
#
# In the application context of solving IBVPs with explicit time integration, the option to have the function value (F) as an unknown
# is useful with Neumann BCs. The neighborhoods of the Neumann boundary points can be chosen such that each Neumann boundary point
# only uses neighbors from the interior of the domain. This gives the possibility to leave F free at all Neumann boundary points,
# while prescribing only a derivative (the normal-direction derivative).
#
# (In practice, at slanted (i.e. not coordinate axis aligned) boundaries, local (tangent, normal)
#  coordinates must be used; i.e., the coordinate system in which the derivatives are to be computed
#  must be rotated to match the orientation of the boundary. This makes Y the normal derivative,
#  which can then be prescribed using this mechanism, while leaving the function value F free.)

# 1D case
#
cdef long long b1_F_c      = (1LL << i1_F_c)
cdef long long b1_X_c      = (1LL << i1_X_c)
cdef long long b1_X2_c     = (1LL << i1_X2_c)
cdef long long b1_X3_c     = (1LL << i1_X3_c)
cdef long long b1_X4_c     = (1LL << i1_X4_c)

# 2D case
#
cdef long long b2_F_c      = (1LL << i2_F_c)

cdef long long b2_X_c      = (1LL << i2_X_c)
cdef long long b2_Y_c      = (1LL << i2_Y_c)

cdef long long b2_X2_c     = (1LL << i2_X2_c)
cdef long long b2_XY_c     = (1LL << i2_XY_c)
cdef long long b2_Y2_c     = (1LL << i2_Y2_c)

cdef long long b2_X3_c     = (1LL << i2_X3_c)
cdef long long b2_X2Y_c    = (1LL << i2_X2Y_c)
cdef long long b2_XY2_c    = (1LL << i2_XY2_c)
cdef long long b2_Y3_c     = (1LL << i2_Y3_c)

cdef long long b2_X4_c     = (1LL << i2_X4_c)
cdef long long b2_X3Y_c    = (1LL << i2_X3Y_c)
cdef long long b2_X2Y2_c   = (1LL << i2_X2Y2_c)
cdef long long b2_XY3_c    = (1LL << i2_XY3_c)
cdef long long b2_Y4_c     = (1LL << i2_Y4_c)

# 3D case
#
cdef long long b3_F_c      = (1LL << i3_F_c)

cdef long long b3_X_c      = (1LL << i3_X_c)
cdef long long b3_Y_c      = (1LL << i3_Y_c)
cdef long long b3_Z_c      = (1LL << i3_Z_c)

cdef long long b3_X2_c     = (1LL << i3_X2_c)
cdef long long b3_XY_c     = (1LL << i3_XY_c)
cdef long long b3_Y2_c     = (1LL << i3_Y2_c)
cdef long long b3_YZ_c     = (1LL << i3_YZ_c)
cdef long long b3_Z2_c     = (1LL << i3_Z2_c)
cdef long long b3_XZ_c     = (1LL << i3_XZ_c)

cdef long long b3_X3_c     = (1LL << i3_X3_c)
cdef long long b3_X2Y_c    = (1LL << i3_X2Y_c)
cdef long long b3_XY2_c    = (1LL << i3_XY2_c)
cdef long long b3_Y3_c     = (1LL << i3_Y3_c)
cdef long long b3_Y2Z_c    = (1LL << i3_Y2Z_c)
cdef long long b3_YZ2_c    = (1LL << i3_YZ2_c)
cdef long long b3_Z3_c     = (1LL << i3_Z3_c)
cdef long long b3_XZ2_c    = (1LL << i3_XZ2_c)
cdef long long b3_X2Z_c    = (1LL << i3_X2Z_c)
cdef long long b3_XYZ_c    = (1LL << i3_XYZ_c)

cdef long long b3_X4_c     = (1LL << i3_X4_c)
cdef long long b3_X3Y_c    = (1LL << i3_X3Y_c)
cdef long long b3_X2Y2_c   = (1LL << i3_X2Y2_c)
cdef long long b3_XY3_c    = (1LL << i3_XY3_c)
cdef long long b3_Y4_c     = (1LL << i3_Y4_c)
cdef long long b3_Y3Z_c    = (1LL << i3_Y3Z_c)
cdef long long b3_Y2Z2_c   = (1LL << i3_Y2Z2_c)
cdef long long b3_YZ3_c    = (1LL << i3_YZ3_c)
cdef long long b3_Z4_c     = (1LL << i3_Z4_c)
cdef long long b3_XZ3_c    = (1LL << i3_XZ3_c)
cdef long long b3_X2Z2_c   = (1LL << i3_X2Z2_c)
cdef long long b3_X3Z_c    = (1LL << i3_X3Z_c)
cdef long long b3_X2YZ_c   = (1LL << i3_X2YZ_c)
cdef long long b3_XY2Z_c   = (1LL << i3_XY2Z_c)
cdef long long b3_XYZ2_c   = (1LL << i3_XYZ2_c)


#################################################
# Python wrapper
#################################################

# Algorithms for the solve step (expert mode).
#
ALGO_BASIC     = ALGO_BASIC_c
ALGO_ITERATIVE = ALGO_ITERATIVE_c

# Weighting methods.
#
WEIGHT_UNIFORM = WEIGHT_UNIFORM_c
WEIGHT_CENTER  = WEIGHT_CENTER_c

# DOF index in the array f.
#
# These are ordered in increasing order of number of differentiations, so that if only first derivatives
# are required, the DOF array can be simply truncated after the first derivatives.
#
# To avoid gaps in the numbering, this requires separate DOF orderings for the 1D, 2D and 3D cases.
#
# (The other logical possibility would be function value first, then x-related, then y-related, then mixed,
#  but then the case of "first derivatives only" requires changes to the ordering to avoid gaps.
#  Specifying different orderings for different numbers of space dimensions is conceptually cleaner
#  of the two possibilities.)

# 1D case
#
i1_F  = i1_F_c
i1_X  = i1_X_c
i1_X2 = i1_X2_c
i1_X3 = i1_X3_c
i1_X4 = i1_X4_c

i1_1st_end = i1_1st_end_c
i1_2nd_end = i1_2nd_end_c
i1_3rd_end = i1_3rd_end_c
i1_4th_end = i1_4th_end_c

SIZE1 = SIZE1_c


# 2D case
#
i2_F     = i2_F_c

i2_X     = i2_X_c
i2_Y     = i2_Y_c

i2_X2    = i2_X2_c
i2_XY    = i2_XY_c
i2_Y2    = i2_Y2_c

i2_X3    = i2_X3_c
i2_X2Y   = i2_X2Y_c
i2_XY2   = i2_XY2_c
i2_Y3    = i2_Y3_c

i2_X4    = i2_X4_c
i2_X3Y   = i2_X3Y_c
i2_X2Y2  = i2_X2Y2_c
i2_XY3   = i2_XY3_c
i2_Y4    = i2_Y4_c

i2_1st_end = i2_1st_end_c
i2_2nd_end = i2_2nd_end_c
i2_3rd_end = i2_3rd_end_c
i2_4th_end = i2_4th_end_c

SIZE2 = SIZE2_c


# 3D case
#
i3_F      = i3_F_c

i3_X      = i3_X_c
i3_Y      = i3_Y_c
i3_Z      = i3_Z_c

i3_X2     = i3_X2_c
i3_XY     = i3_XY_c
i3_Y2     = i3_Y2_c
i3_YZ     = i3_YZ_c
i3_Z2     = i3_Z2_c
i3_XZ     = i3_XZ_c

i3_X3     = i3_X3_c
i3_X2Y    = i3_X2Y_c
i3_XY2    = i3_XY2_c
i3_Y3     = i3_Y3_c
i3_Y2Z    = i3_Y2Z_c
i3_YZ2    = i3_YZ2_c
i3_Z3     = i3_Z3_c
i3_XZ2    = i3_XZ2_c
i3_X2Z    = i3_X2Z_c
i3_XYZ    = i3_XYZ_c

i3_X4     = i3_X4_c
i3_X3Y    = i3_X3Y_c
i3_X2Y2   = i3_X2Y2_c
i3_XY3    = i3_XY3_c
i3_Y4     = i3_Y4_c
i3_Y3Z    = i3_Y3Z_c
i3_Y2Z2   = i3_Y2Z2_c
i3_YZ3    = i3_YZ3_c
i3_Z4     = i3_Z4_c
i3_XZ3    = i3_XZ3_c
i3_X2Z2   = i3_X2Z2_c
i3_X3Z    = i3_X3Z_c
i3_X2YZ   = i3_X2YZ_c
i3_XY2Z   = i3_XY2Z_c
i3_XYZ2   = i3_XYZ2_c

i3_0th_end = i3_0th_end_c
i3_1st_end = i3_1st_end_c
i3_2nd_end = i3_2nd_end_c
i3_3rd_end = i3_3rd_end_c
i3_4th_end = i3_4th_end_c

SIZE3 = SIZE3_c


# bitmask constants for knowns.
#
# Knowns are eliminated algebraically from the equation system; if any knowns are specified,
# the system to be solved (for a point x_i) will be smaller than the full 6x6.
#
# The sensible default thing to do is to consider the function value F known, with all the
# derivatives unknown.
#
# Note that here "known" means "known at point x_i" (the point at which we wish to compute the derivatives).
#
# Function values (F) are always assumed known at all *neighbor* points x_k, since they are used
# for determining the local least-squares quadratic polynomial fit to the data. This fit is then used
# as local a surrogate model for the unknown function f; in WLSQM, the derivatives are actually computed
# from the surrogate.
#
# The option to have the function value (F) as an unknown is useful with Neumann BCs, if the neighborhoods
# of the Neumann boundary points are chosen so that each Neumann boundary point only uses neighbors from
# the interior of the domain. This gives the possibility to leave F free at all Neumann boundary points,
# while prescribing only a derivative.
#
# (In practice, at slanted (i.e. not coordinate axis aligned) boundaries, local (tangent, normal)
#  coordinates must be used; i.e., the coordinate system in which the derivatives are to be computed
#  must be rotated to match the orientation of the boundary. This makes Y the normal derivative,
#  which can then be prescribed using this mechanism, while leaving the function value F free.)

# 1D case
#
b1_F  = b1_F_c
b1_X  = b1_X_c
b1_X2 = b1_X2_c
b1_X3 = b1_X3_c
b1_X4 = b1_X4_c


# 2D case
#
b2_F     = b2_F_c

b2_X     = b2_X_c
b2_Y     = b2_Y_c

b2_X2    = b2_X2_c
b2_XY    = b2_XY_c
b2_Y2    = b2_Y2_c

b2_X3    = b2_X3_c
b2_X2Y   = b2_X2Y_c
b2_XY2   = b2_XY2_c
b2_Y3    = b2_Y3_c

b2_X4    = b2_X4_c
b2_X3Y   = b2_X3Y_c
b2_X2Y2  = b2_X2Y2_c
b2_XY3   = b2_XY3_c
b2_Y4    = b2_Y4_c


# 3D case
#
b3_F      = b3_F_c

b3_X      = b3_X_c
b3_Y      = b3_Y_c
b3_Z      = b3_Z_c

b3_X2     = b3_X2_c
b3_XY     = b3_XY_c
b3_Y2     = b3_Y2_c
b3_YZ     = b3_YZ_c
b3_Z2     = b3_Z2_c
b3_XZ     = b3_XZ_c

b3_X3     = b3_X3_c
b3_X2Y    = b3_X2Y_c
b3_XY2    = b3_XY2_c
b3_Y3     = b3_Y3_c
b3_Y2Z    = b3_Y2Z_c
b3_YZ2    = b3_YZ2_c
b3_Z3     = b3_Z3_c
b3_XZ2    = b3_XZ2_c
b3_X2Z    = b3_X2Z_c
b3_XYZ    = b3_XYZ_c

b3_X4     = b3_X4_c
b3_X3Y    = b3_X3Y_c
b3_X2Y2   = b3_X2Y2_c
b3_XY3    = b3_XY3_c
b3_Y4     = b3_Y4_c
b3_Y3Z    = b3_Y3Z_c
b3_Y2Z2   = b3_Y2Z2_c
b3_YZ3    = b3_YZ3_c
b3_Z4     = b3_Z4_c
b3_XZ3    = b3_XZ3_c
b3_X2Z2   = b3_X2Z2_c
b3_X3Z    = b3_X3Z_c
b3_X2YZ   = b3_X2YZ_c
b3_XY2Z   = b3_XY2Z_c
b3_XYZ2   = b3_XYZ2_c

